import logging
from datetime import datetime, timedelta
from types import SimpleNamespace
from typing import List, Optional, Tuple

from sqlalchemy import and_, or_
from sqlalchemy.orm import Query
from typing_extensions import Literal

from cg.constants import CASE_ACTIONS, Pipeline
from cg.store import models
from cg.store.api.base import BaseHandler
from cg.utils.date import get_date

VALID_DATA_IN_PRODUCTION = get_date("2017-09-27")


class StatusHandler(BaseHandler):
    """Handles status states for entities in the database"""

    def samples_to_receive(self, external=False) -> Query:
        """Fetch incoming samples."""
        return (
            self.Sample.query.join(
                models.Sample.application_version,
                models.ApplicationVersion.application,
            )
            .filter(
                models.Sample.received_at.is_(None),
                models.Sample.downsampled_to.is_(None),
                models.Application.is_external == external,
            )
            .order_by(models.Sample.ordered_at)
        )

    def samples_to_prepare(self) -> Query:
        """Fetch samples in lab prep queue."""
        return (
            self.Sample.query.join(
                models.Sample.application_version,
                models.ApplicationVersion.application,
            )
            .filter(
                models.Sample.received_at.isnot(None),
                models.Sample.prepared_at.is_(None),
                models.Sample.downsampled_to.is_(None),
                models.Application.is_external == False,
                models.Sample.sequenced_at.is_(None),
            )
            .order_by(models.Sample.received_at)
        )

    def samples_to_sequence(self) -> Query:
        """Fetch samples in sequencing."""
        return (
            self.Sample.query.join(
                models.Sample.application_version,
                models.ApplicationVersion.application,
            )
            .filter(
                models.Sample.prepared_at.isnot(None),
                models.Sample.sequenced_at.is_(None),
                models.Sample.downsampled_to.is_(None),
                models.Application.is_external == False,
            )
            .order_by(models.Sample.received_at)
        )

    def cases_to_analyze(
        self, pipeline: Pipeline = None, threshold: bool = False, limit: int = None
    ) -> List[models.Family]:
        """Returns a list if cases ready to be analyzed or set to be reanalyzed"""
        families_query = list(
            self.Family.query.outerjoin(models.Analysis)
            .join(
                models.Family.links,
                models.FamilySample.sample,
                models.ApplicationVersion,
                models.Application,
            )
            .filter(or_(models.Application.is_external, models.Sample.sequenced_at.isnot(None)))
            .filter(models.Family.data_analysis == str(pipeline))
            .filter(
                or_(
                    models.Family.action == "analyze",
                    and_(
                        models.Application.is_external.isnot(True),
                        models.Family.action.is_(None),
                        models.Analysis.created_at.is_(None),
                    ),
                    and_(
                        models.Family.action.is_(None),
                        models.Analysis.created_at < models.Sample.sequenced_at,
                    ),
                )
            )
            .order_by(models.Family.ordered_at)
        )
        families_query = [
            case_obj
            for case_obj in families_query
            if case_obj.latest_sequenced
            and (
                case_obj.action == "analyze"
                or not case_obj.latest_analyzed
                or case_obj.latest_analyzed < case_obj.latest_sequenced
            )
        ]

        if threshold:
            families_query = [
                case_obj for case_obj in families_query if case_obj.all_samples_pass_qc
            ]
        return families_query[:limit]

    def cases_to_store(self, pipeline: Pipeline, limit: int = None) -> list:
        """Returns a list of cases that may be available to store in Housekeeper"""
        families_query = (
            self.Family.query.outerjoin(models.Analysis)
            .join(models.Family.links, models.FamilySample.sample)
            .filter(models.Family.data_analysis == str(pipeline))
            .filter(models.Family.action == "running")
        )
        return list(families_query)[:limit]

    def get_running_cases_for_pipeline(self, pipeline: Pipeline) -> List[models.Family]:
        return (
            self.query(models.Family)
            .filter(models.Family.action == "running")
            .filter(models.Family.data_analysis == pipeline)
            .all()
        )

    def cases(
        self,
        internal_id: str = None,
        name: str = None,
        days: int = 0,
        case_action: Optional[str] = None,
        priority: str = None,
        customer_id: str = None,
        exclude_customer_id: str = None,
        data_analysis: str = None,
        sample_id: str = None,
        only_received: bool = False,
        only_prepared: bool = False,
        only_sequenced: bool = False,
        only_analysed: bool = False,
        only_uploaded: bool = False,
        only_delivered: bool = False,
        only_delivery_reported: bool = False,
        only_invoiced: bool = False,
        exclude_received: bool = False,
        exclude_prepared: bool = False,
        exclude_sequenced: bool = False,
        exclude_analysed: bool = False,
        exclude_uploaded: bool = False,
        exclude_delivered: bool = False,
        exclude_delivery_reported: bool = False,
        exclude_invoiced: bool = False,
    ) -> List[models.Family]:
        """Fetch cases with and w/o analyses"""
        case_q = self._get_filtered_case_query(
            case_action,
            customer_id,
            data_analysis,
            days,
            exclude_customer_id,
            internal_id,
            name,
            priority,
            sample_id,
        )

        cases = []

        for case_obj in case_q:

            case_data = self._calculate_case_data(case_obj)

            skip_case = self._should_be_skipped(
                case_data,
                exclude_analysed,
                exclude_delivered,
                exclude_delivery_reported,
                exclude_invoiced,
                exclude_prepared,
                exclude_received,
                exclude_sequenced,
                exclude_uploaded,
                only_analysed,
                only_delivered,
                only_delivery_reported,
                only_invoiced,
                only_prepared,
                only_received,
                only_sequenced,
                only_uploaded,
            )

            if skip_case:
                continue

            case_output = self._get_case_output(case_data)

            cases.append(case_output)

        return sorted(cases, key=lambda k: k["tat"], reverse=True)

    def set_case_action(self, action: Literal[CASE_ACTIONS], case_id: str) -> None:
        """Sets the action of provided cases to None or the given action"""
        case_obj: models.Family = self.Family.query.filter(
            models.Family.internal_id == case_id
        ).first()
        case_obj.action = action
        self.commit()

    @staticmethod
    def _get_case_output(case_data: SimpleNamespace) -> dict:
        return {
            "data_analysis": case_data.data_analysis,
            "internal_id": case_data.internal_id,
            "name": case_data.name,
            "ordered_at": case_data.ordered_at,
            "total_samples": case_data.total_samples,
            "total_external_samples": case_data.total_external_samples,
            "total_internal_samples": case_data.total_internal_samples,
            "case_external_bool": case_data.case_external_bool,
            "samples_to_receive": case_data.samples_to_receive,
            "samples_to_prepare": case_data.samples_to_prepare,
            "samples_to_sequence": case_data.samples_to_sequence,
            "samples_to_deliver": case_data.samples_to_deliver,
            "samples_to_invoice": case_data.samples_to_invoice,
            "samples_received": case_data.samples_received,
            "samples_prepared": case_data.samples_prepared,
            "samples_sequenced": case_data.samples_sequenced,
            "samples_received_at": case_data.samples_received_at,
            "samples_prepared_at": case_data.samples_prepared_at,
            "samples_sequenced_at": case_data.samples_sequenced_at,
            "samples_delivered_at": case_data.samples_delivered_at,
            "samples_invoiced_at": case_data.samples_invoiced_at,
            "case_action": case_data.case_action,
            "analysis_completed_at": case_data.analysis_completed_at,
            "analysis_uploaded_at": case_data.analysis_uploaded_at,
            "samples_delivered": case_data.samples_delivered,
            "analysis_delivery_reported_at": case_data.analysis_delivery_reported_at,
            "samples_invoiced": case_data.samples_invoiced,
            "analysis_pipeline": case_data.analysis_pipeline,
            "samples_received_bool": case_data.samples_received_bool,
            "samples_prepared_bool": case_data.samples_prepared_bool,
            "samples_sequenced_bool": case_data.samples_sequenced_bool,
            "analysis_completed_bool": case_data.analysis_completed_bool,
            "analysis_uploaded_bool": case_data.analysis_uploaded_bool,
            "samples_delivered_bool": case_data.samples_delivered_bool,
            "analysis_delivery_reported_bool": case_data.analysis_delivery_reported_bool,
            "samples_invoiced_bool": case_data.samples_invoiced_bool,
            "flowcells_status": case_data.flowcells_status,
            "flowcells_on_disk": case_data.flowcells_on_disk,
            "flowcells_on_disk_bool": case_data.flowcells_on_disk_bool,
            "tat": case_data.tat,
            "is_rerun": case_data.is_rerun,
            "max_tat": case_data.max_tat,
        }

    @staticmethod
    def _should_be_skipped(
        case_data: SimpleNamespace,
        exclude_analysed: bool,
        exclude_delivered: bool,
        exclude_delivery_reported: bool,
        exclude_invoiced: bool,
        exclude_prepared: bool,
        exclude_received: bool,
        exclude_sequenced: bool,
        exclude_uploaded: bool,
        only_analysed: bool,
        only_delivered: bool,
        only_delivery_reported: bool,
        only_invoiced: bool,
        only_prepared: bool,
        only_received: bool,
        only_sequenced: bool,
        only_uploaded: bool,
    ) -> bool:
        skip_case = False
        if only_received and not case_data.samples_received_bool:
            skip_case = True
        if only_prepared and not case_data.samples_prepared_bool:
            skip_case = True
        if only_sequenced and not case_data.samples_sequenced_bool:
            skip_case = True
        if only_analysed and not case_data.analysis_completed_bool:
            skip_case = True
        if only_uploaded and not case_data.analysis_uploaded_bool:
            skip_case = True
        if only_delivered and not case_data.samples_delivered_bool:
            skip_case = True
        if only_delivery_reported and not case_data.analysis_delivery_reported_bool:
            skip_case = True
        if only_invoiced and not case_data.samples_invoiced_bool:
            skip_case = True
        if exclude_received and case_data.samples_received_bool:
            skip_case = True
        if exclude_prepared and case_data.samples_prepared_bool:
            skip_case = True
        if exclude_sequenced and case_data.samples_sequenced_bool:
            skip_case = True
        if exclude_analysed and case_data.analysis_completed_bool:
            skip_case = True
        if exclude_uploaded and case_data.analysis_uploaded_bool:
            skip_case = True
        if exclude_delivered and case_data.samples_delivered_bool:
            skip_case = True
        if exclude_delivery_reported and case_data.analysis_delivery_reported_bool:
            skip_case = True
        if exclude_invoiced and case_data.samples_invoiced_bool:
            skip_case = True
        return skip_case

    def _calculate_case_data(self, case_obj: models.Family) -> SimpleNamespace:
        case_data = self._get_empty_case_data()

        case_data.data_analysis = case_obj.data_analysis
        case_data.internal_id = case_obj.internal_id
        case_data.name = case_obj.name
        case_data.ordered_at = case_obj.ordered_at

        case_data.analysis_in_progress = case_obj.action == "analyze"
        case_data.case_action = case_obj.action
        case_data.total_samples = len(case_obj.links)
        case_data.total_external_samples = len(
            [
                link.sample.application_version.application.is_external
                for link in case_obj.links
                if link.sample.application_version.application.is_external
            ]
        )
        case_data.total_internal_samples = (
            case_data.total_samples - case_data.total_external_samples
        )
        case_data.case_external_bool = case_data.total_external_samples == case_data.total_samples
        if case_data.total_samples > 0:
            case_data.samples_received = len(
                [link.sample.received_at for link in case_obj.links if link.sample.received_at]
            )
            case_data.samples_prepared = len(
                [link.sample.prepared_at for link in case_obj.links if link.sample.prepared_at]
            )
            case_data.samples_sequenced = len(
                [link.sample.sequenced_at for link in case_obj.links if link.sample.sequenced_at]
            )
            case_data.samples_delivered = len(
                [link.sample.delivered_at for link in case_obj.links if link.sample.delivered_at]
            )
            case_data.samples_invoiced = len(
                [
                    link.sample.invoice.invoiced_at
                    for link in case_obj.links
                    if link.sample.invoice and link.sample.invoice.invoiced_at
                ]
            )

            case_data.samples_to_receive = case_data.total_internal_samples
            case_data.samples_to_prepare = case_data.total_internal_samples
            case_data.samples_to_sequence = case_data.total_internal_samples
            case_data.samples_to_deliver = case_data.total_internal_samples
            case_data.samples_to_invoice = case_data.total_samples - len(
                [link.sample.no_invoice for link in case_obj.links if link.sample.no_invoice]
            )

            case_data.samples_received_bool = (
                case_data.samples_received == case_data.samples_to_receive
            )
            case_data.samples_prepared_bool = (
                case_data.samples_prepared == case_data.samples_to_prepare
            )
            case_data.samples_sequenced_bool = (
                case_data.samples_sequenced == case_data.samples_to_sequence
            )
            case_data.samples_delivered_bool = (
                case_data.samples_delivered == case_data.samples_to_deliver
            )
            case_data.samples_invoiced_bool = (
                case_data.samples_invoiced == case_data.samples_to_invoice
            )

            if case_data.samples_to_receive > 0 and case_data.samples_received_bool:
                case_data.samples_received_at = max(
                    link.sample.received_at
                    for link in case_obj.links
                    if link.sample.received_at is not None
                )

            if case_data.samples_to_prepare > 0 and case_data.samples_prepared_bool:
                case_data.samples_prepared_at = max(
                    link.sample.prepared_at
                    for link in case_obj.links
                    if link.sample.prepared_at is not None
                )

            if case_data.samples_to_sequence > 0 and case_data.samples_sequenced_bool:
                case_data.samples_sequenced_at = max(
                    link.sample.sequenced_at
                    for link in case_obj.links
                    if link.sample.sequenced_at is not None
                )

            if case_data.samples_to_deliver > 0 and case_data.samples_delivered_bool:
                case_data.samples_delivered_at = max(
                    link.sample.delivered_at
                    for link in case_obj.links
                    if link.sample.delivered_at is not None
                )

            if case_data.samples_to_invoice > 0 and case_data.samples_invoiced_bool:
                case_data.samples_invoiced_at = max(
                    link.sample.invoice.invoiced_at
                    for link in case_obj.links
                    if link.sample.invoice and link.sample.invoice.invoiced_at
                )

            case_data.flowcells = len(
                [flowcell.status for link in case_obj.links for flowcell in link.sample.flowcells]
            )

            case_data.flowcells_status = list(
                {flowcell.status for link in case_obj.links for flowcell in link.sample.flowcells}
            )

            if case_data.flowcells < case_data.total_samples:
                case_data.flowcells_status.append("new")

            case_data.flowcells_status = ", ".join(case_data.flowcells_status)

            case_data.flowcells_on_disk = len(
                [
                    flowcell.status
                    for link in case_obj.links
                    for flowcell in link.sample.flowcells
                    if flowcell.status == "ondisk"
                ]
            )

            case_data.flowcells_on_disk_bool = (
                case_data.flowcells_on_disk == case_data.total_samples
            )
        if case_obj.analyses and not case_data.analysis_in_progress:
            case_data.analysis_completed_at = case_obj.analyses[0].completed_at
            case_data.analysis_uploaded_at = case_obj.analyses[0].uploaded_at
            case_data.analysis_delivery_reported_at = case_obj.analyses[
                0
            ].delivery_report_created_at
            case_data.analysis_pipeline = case_obj.analyses[0].pipeline
            case_data.analysis_completed_bool = case_data.analysis_completed_at is not None
            case_data.analysis_uploaded_bool = case_data.analysis_uploaded_at is not None
            case_data.analysis_delivery_reported_bool = (
                case_data.analysis_delivery_reported_at is not None
            )
        elif case_data.total_samples > 0:
            case_data.analysis_completed_bool = False
            case_data.analysis_uploaded_bool = False
            case_data.analysis_delivery_reported_bool = False

        case_data.is_rerun = self._is_rerun(
            case_obj,
            case_data.samples_received_at,
            case_data.samples_prepared_at,
            case_data.samples_sequenced_at,
        )
        case_data.tat = self._calculate_estimated_turnaround_time(
            case_data.is_rerun,
            case_data.case_external_bool,
            case_obj.ordered_at,
            case_data.samples_received_at,
            case_data.samples_prepared_at,
            case_data.samples_sequenced_at,
            case_data.analysis_completed_at,
            case_data.analysis_uploaded_at,
            case_data.samples_delivered_at,
        )
        case_data.max_tat = self._get_max_tat(links=case_obj.links)
        return case_data

    def _get_filtered_case_query(
        self,
        case_action: Optional[str],
        customer_id: str,
        data_analysis: str,
        days: int,
        exclude_customer_id: bool,
        internal_id: str,
        name: str,
        priority: str,
        sample_id: str,
    ) -> Query:
        case_q = self.Family.query
        # family filters
        if days != 0:
            filter_date = datetime.now() - timedelta(days=days)
            case_q = case_q.filter(models.Family.ordered_at > filter_date)
        if case_action:
            case_q = case_q.filter(models.Family.action == case_action)
        if priority:
            case_q = case_q.filter(models.Family.priority == priority)
        if internal_id:
            case_q = case_q.filter(models.Family.internal_id.ilike(f"%{internal_id}%"))
        if name:
            case_q = case_q.filter(models.Family.name.ilike(f"%{name}%"))
        if data_analysis:
            case_q = case_q.filter(models.Family.data_analysis.ilike(f"%{data_analysis}%"))
        # customer filters
        if customer_id or exclude_customer_id:
            case_q = case_q.join(models.Family.customer)

        if customer_id:
            case_q = case_q.filter(models.Customer.internal_id == customer_id)

        if exclude_customer_id:
            case_q = case_q.filter(models.Customer.internal_id != exclude_customer_id)
        # sample filters
        if sample_id:
            case_q = case_q.join(models.Family.links, models.FamilySample.sample)
            case_q = case_q.filter(models.Sample.internal_id.ilike(f"%{sample_id}%"))
        else:
            case_q = case_q.outerjoin(models.Family.links, models.FamilySample.sample)
        # other joins
        case_q = case_q.outerjoin(
            models.Family.analyses, models.Sample.invoice, models.Sample.flowcells
        )
        return case_q

    @staticmethod
    def _is_rerun(
        case_obj: models.Family,
        samples_received_at: datetime,
        samples_prepared_at: datetime,
        samples_sequenced_at: datetime,
    ) -> bool:

        return (
            (len(case_obj.analyses) > 0)
            or (samples_received_at and samples_received_at < case_obj.ordered_at)
            or (samples_prepared_at and samples_prepared_at < case_obj.ordered_at)
            or (samples_sequenced_at and samples_sequenced_at < case_obj.ordered_at)
        )

    @staticmethod
    def _all_samples_have_sequence_data(links: List[models.FamilySample]) -> bool:
        """Return True if all samples are external or sequenced in-house."""
        return all(
            (link.sample.sequenced_at or link.sample.application_version.application.is_external)
            for link in links
        )

    def analyses_to_upload(self, pipeline: Pipeline = None) -> List[models.Analysis]:
        """Fetch analyses that haven't been uploaded."""
        records = self.Analysis.query.filter(
            models.Analysis.completed_at.isnot(None), models.Analysis.uploaded_at.is_(None)
        )

        if pipeline:
            records = records.filter(models.Analysis.pipeline == str(pipeline))

        return records

    def analyses_to_clean(
        self, before: datetime = datetime.now(), pipeline: Pipeline = None
    ) -> Query:
        """Fetch analyses that haven't been cleaned."""
        records = self.latest_analyses()
        records = records.filter(
            models.Analysis.uploaded_at.isnot(None),
            models.Analysis.cleaned_at.is_(None),
            models.Analysis.started_at <= before,
            models.Family.action.is_(None),
        )
        if pipeline:
            records = records.filter(
                models.Analysis.pipeline == str(pipeline),
            )

        return records

    def get_analyses_before_date(
        self,
        case_id: Optional[str] = None,
        before: Optional[datetime] = datetime.now(),
        pipeline: Optional[Pipeline] = None,
    ) -> Query:
        """Fetch all analyses older than certain date"""
        records = self.Analysis.query.join(models.Analysis.family)
        if case_id:
            records = records.filter(models.Family.internal_id == case_id)
        if pipeline:
            records = records.filter(
                models.Analysis.pipeline == str(pipeline),
            )
        records = records.filter(models.Analysis.started_at <= before)
        return records

    def observations_to_upload(self):
        """Fetch observations that haven't been uploaded."""

        return self.Family.query.join(
            models.Analysis, models.Family.links, models.FamilySample.sample
        ).filter(models.Sample.loqusdb_id.is_(None))

    def observations_uploaded(self) -> Query:
        """Fetch observations that have been uploaded."""

        return self.Family.query.join(models.Family.links, models.FamilySample.sample).filter(
            models.Sample.loqusdb_id.isnot(None)
        )

    def analyses_to_deliver(self, pipeline: Pipeline = None) -> Query:
        """Fetch analyses that have been uploaded but not delivered."""
        return (
            self.Analysis.query.join(models.Family, models.Family.links, models.FamilySample.sample)
            .filter(
                models.Analysis.uploaded_at.isnot(None),
                models.Sample.delivered_at.is_(None),
                models.Analysis.pipeline == str(pipeline),
            )
            .order_by(models.Analysis.uploaded_at.desc())
        )

    def analyses_to_delivery_report(self, pipeline: Pipeline = None) -> Query:
        """Fetch analyses that needs the delivery report to be regenerated."""

        analyses_query = self.latest_analyses()

        analyses_query = (
            analyses_query.filter(models.Analysis.uploaded_at)
            .filter(VALID_DATA_IN_PRODUCTION < models.Analysis.started_at)
            .join(models.Family, models.Family.links, models.FamilySample.sample)
            .filter(
                or_(
                    models.Family.data_analysis.is_(None),
                    models.Family.data_analysis == str(pipeline),
                )
            )
            .filter(
                models.Sample.delivered_at.isnot(None),
                or_(
                    models.Analysis.delivery_report_created_at.is_(None),
                    and_(
                        models.Analysis.delivery_report_created_at.isnot(None),
                        models.Analysis.delivery_report_created_at < models.Sample.delivered_at,
                    ),
                ),
            )
            .order_by(models.Analysis.uploaded_at.desc())
        )
        return analyses_query

    def samples_to_deliver(self) -> Query:
        """Fetch samples that have been sequenced but not delivered."""
        return self.Sample.query.filter(
            models.Sample.sequenced_at.isnot(None),
            models.Sample.delivered_at.is_(None),
            models.Sample.downsampled_to.is_(None),
        )

    def samples_not_delivered(self) -> Query:
        """Fetch samples not delivered."""
        return self.Sample.query.filter(
            models.Sample.delivered_at.is_(None), models.Sample.downsampled_to.is_(None)
        )

    def samples_not_invoiced(self) -> Query:
        """Fetch all samples that are not invoiced."""
        return self.Sample.query.filter(
            models.Sample.downsampled_to.is_(None), models.Sample.invoice_id.is_(None)
        )

    def samples_not_downsampled(self) -> Query:
        """Fetch all samples that are not down sampled."""
        return self.Sample.query.filter(models.Sample.downsampled_to.is_(None))

    def microbial_samples_to_invoice(self, customer: models.Customer = None) -> Tuple[Query, list]:
        """Fetch microbial samples that should be invoiced.

        Returns microbial samples that have been delivered but not invoiced.
        """
        records = self.Sample.query.filter(
            str(Pipeline.MICROSALT) in self.Family.data_analysis,
            models.Sample.delivered_at is not None,
            models.Sample.invoice_id.is_(None),
        )
        customers_to_invoice = list({case_obj.customer for case_obj in records.all()})
        if customer:
            records = records.join(models.Family).filter(models.Family.customer_id == customer.id)
        return records, customers_to_invoice

    def samples_to_invoice(self, customer: models.Customer = None) -> Tuple[Query, list]:
        """Fetch samples that should be invoiced.

        Returns samples have been delivered but not invoiced, excluding those that
        have been marked to skip invoicing.
        """
        records = self.Sample.query.filter(
            models.Sample.delivered_at.isnot(None),
            models.Sample.invoice_id.is_(None),
            models.Sample.no_invoice == False,
            models.Sample.downsampled_to.is_(None),
        )
        customers_to_invoice = [
            case_obj.customer
            for case_obj in records.all()
            if case_obj.customer.internal_id != "cust000"
        ]

        customers_to_invoice = list(set(customers_to_invoice))
        records = records.filter(models.Sample.customer == customer) if customer else records
        return records, customers_to_invoice

    def pools_to_invoice(self, customer: models.Customer = None) -> Tuple[Query, list]:
        """
        Fetch pools that should be invoiced.
        """
        records = self.Pool.query.filter(
            models.Pool.invoice_id.is_(None),
            models.Pool.no_invoice == False,
            models.Pool.delivered_at.isnot(None),
        )

        customers_to_invoice = [
            case_obj.customer
            for case_obj in records.all()
            if case_obj.customer.internal_id != "cust000"
        ]

        customers_to_invoice = list(set(customers_to_invoice))
        records = records.filter(models.Pool.customer_id == customer.id) if customer else records
        return records, customers_to_invoice

    def pools_to_receive(self) -> Query:
        """Fetch pools that have been not yet been received."""
        return self.Pool.query.filter(models.Pool.received_at.is_(None))

    def pools_to_deliver(self) -> Query:
        """Fetch pools that have been not yet been delivered."""
        return self.Pool.query.filter(
            models.Pool.received_at.isnot(None), models.Pool.delivered_at.is_(None)
        )

    def _calculate_estimated_turnaround_time(
        self,
        is_rerun,
        external_case_bool,
        analysis_ordered_at,
        samples_received_at,
        samples_prepared_at,
        samples_sequenced_at,
        analysis_completed_at,
        analysis_uploaded_at,
        samples_delivered_at,
    ) -> timedelta:
        """Calculated estimated turnaround-time"""
        if samples_received_at and samples_delivered_at:
            return self._calculate_date_delta(None, samples_received_at, samples_delivered_at)

        o_a = self._calculate_date_delta(5, analysis_ordered_at, analysis_completed_at)
        r_p = self._calculate_date_delta(4, samples_received_at, samples_prepared_at)
        p_s = self._calculate_date_delta(5, samples_prepared_at, samples_sequenced_at)
        s_a = self._calculate_date_delta(4, samples_sequenced_at, analysis_completed_at)
        a_u = self._calculate_date_delta(1, analysis_completed_at, analysis_uploaded_at)
        u_d = self._calculate_date_delta(2, analysis_uploaded_at, samples_delivered_at)

        if is_rerun:
            o_a = self._calculate_date_delta(1, analysis_ordered_at, analysis_completed_at)
            return o_a + a_u

        if external_case_bool:
            if analysis_ordered_at and analysis_uploaded_at:
                return self._calculate_date_delta(None, analysis_ordered_at, analysis_uploaded_at)

            return o_a + a_u

        if samples_received_at and samples_delivered_at:
            return self._calculate_date_delta(None, samples_received_at, samples_delivered_at)

        return r_p + p_s + s_a + a_u + u_d

    @staticmethod
    def _calculate_date_delta(default, first_date, last_date) -> timedelta:
        # calculates date delta between two dates, assumes last_date is today if missing
        delta = default
        if not last_date:
            last_date = datetime.now()
        if first_date:
            delta = (last_date - first_date).days
        return delta

    @staticmethod
    def _get_max_tat(links) -> int:
        max_tat = 0
        for link in links:
            if link.sample.application_version.application.turnaround_time:
                max_tat = max(0, link.sample.application_version.application.turnaround_time)
        return max_tat

    @staticmethod
    def _get_empty_case_data() -> SimpleNamespace:
        case_data = SimpleNamespace()
        case_data.data_analysis = None
        case_data.internal_id = None
        case_data.name = None
        case_data.ordered_at = None
        case_data.total_samples = None
        case_data.total_external_samples = None
        case_data.total_internal_samples = None
        case_data.case_external_bool = None
        case_data.samples_to_receive = None
        case_data.samples_to_prepare = None
        case_data.samples_to_sequence = None
        case_data.samples_to_deliver = None
        case_data.samples_to_invoice = None
        case_data.samples_received = None
        case_data.samples_prepared = None
        case_data.samples_sequenced = None
        case_data.samples_received_at = None
        case_data.samples_prepared_at = None
        case_data.samples_sequenced_at = None
        case_data.samples_delivered_at = None
        case_data.samples_invoiced_at = None
        case_data.case_action = None
        case_data.analysis_completed_at = None
        case_data.analysis_uploaded_at = None
        case_data.samples_delivered = None
        case_data.analysis_delivery_reported_at = None

        case_data.samples_invoiced = None
        case_data.analysis_pipeline = None
        case_data.samples_received_bool = None
        case_data.samples_prepared_bool = None
        case_data.samples_sequenced_bool = None
        case_data.analysis_completed_bool = None
        case_data.analysis_uploaded_bool = None
        case_data.samples_delivered_bool = None
        case_data.analysis_delivery_reported_bool = None
        case_data.samples_invoiced_bool = None
        case_data.flowcells_status = None
        case_data.flowcells_on_disk = None
        case_data.flowcells_on_disk_bool = None
        case_data.tat = None
        case_data.is_rerun = None
        case_data.max_tat = None

        return case_data



from pathlib import Path

import pytest


@pytest.fixture(name="spring_file_path")
def fixture_spring_file_path() -> Path:
    """Return spring file path"""
    return Path("/path/to/spring_file.spring")


@pytest.fixture(name="backup_file_path")
def fixture_backup_file_path() -> str:
    """Return path to a file used in the backup process"""
    return "/path/to/backup_file.extension"



""" Module for retrieving FCs from backup """
import logging
import subprocess
from pathlib import Path
from typing import Optional

from housekeeper.store import models as hk_models

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.constants import FlowCellStatus
from cg.exc import ChecksumFailedError
from cg.meta.backup.pdc import PdcAPI
from cg.meta.encryption.encryption import SpringEncryptionAPI
from cg.models import CompressionData
from cg.store import Store, models
from cg.utils.time import get_elapsed_time, get_start_time

LOG = logging.getLogger(__name__)


class BackupApi:
    """Class for retrieving FCs from backup"""

    def __init__(self, status: Store, pdc_api: PdcAPI, root_dir: dict):

        self.status: Store = status
        self.pdc: PdcAPI = pdc_api
        self.root_dir: dict = root_dir

    def check_processing(self, max_processing_flow_cells: int = 1) -> bool:
        """Check if the processing queue for flow cells is not full."""
        processing_flow_cells = self.status.flowcells(status=FlowCellStatus.PROCESSING).count()
        LOG.debug("Processing flow cells: %s", processing_flow_cells)
        return processing_flow_cells < max_processing_flow_cells

    def get_first_flow_cell(self) -> Optional[models.Flowcell]:
        """Get the first flow cell from the requested queue"""
        flow_cell_obj: Optional[models.Flowcell] = self.status.flowcells(
            status=FlowCellStatus.REQUESTED
        ).first()
        if not flow_cell_obj:
            return None
        return flow_cell_obj

    def fetch_flow_cell(
        self, flow_cell_obj: Optional[models.Flowcell] = None, dry_run: bool = False
    ) -> Optional[float]:
        """Start fetching a flow cell from backup if possible.

        1. The processing queue is not full
        2. The requested queue is not emtpy
        """
        if self.check_processing() is False:
            LOG.info("Processing queue is full")
            return None

        if not flow_cell_obj:
            flow_cell_obj = self.get_first_flow_cell()

        if not flow_cell_obj:
            LOG.info("No flow cells requested")
            return None

        flow_cell_obj.status = FlowCellStatus.PROCESSING
        if not dry_run:
            self.status.commit()
            LOG.info("%s: retrieving from PDC", flow_cell_obj.name)

        start_time = get_start_time()

        try:
            self.pdc.retrieve_flow_cell(
                flow_cell_id=flow_cell_obj.name,
                sequencer_type=flow_cell_obj.sequencer_type,
                root_dir=self.root_dir,
                dry=dry_run,
            )
            if not dry_run:
                flow_cell_obj.status = FlowCellStatus.RETRIEVED
                self.status.commit()
                LOG.info(
                    "Status for flow cell %s set to %s", flow_cell_obj.name, flow_cell_obj.status
                )
        except subprocess.CalledProcessError as error:
            LOG.error("%s: retrieval failed", flow_cell_obj.name)
            if not dry_run:
                flow_cell_obj.status = FlowCellStatus.REQUESTED
                self.status.commit()
            raise error

        return get_elapsed_time(start_time=start_time)


class SpringBackupAPI:
    """Class for handling PDC backup and retrieval of spring compressed fastq files"""

    def __init__(
        self,
        encryption_api: SpringEncryptionAPI,
        pdc_api: PdcAPI,
        hk_api: HousekeeperAPI,
        dry_run: bool = False,
    ):
        self.encryption_api: SpringEncryptionAPI = encryption_api
        self.hk_api: HousekeeperAPI = hk_api
        self.pdc: PdcAPI = pdc_api
        self._dry_run = dry_run

    def encrypt_and_archive_spring_file(self, spring_file_path: Path) -> None:
        """Encrypts and archives a spring file and its decryption key"""
        LOG.debug(f"*** START BACKUP PROCESS OF SPRING FILE %s ***", spring_file_path)
        if self.is_compression_ongoing(spring_file_path):
            LOG.info(
                "Spring (de)compression ongoing, skipping archiving for spring file %s",
                spring_file_path,
            )
            return
        self.encryption_api.cleanup(spring_file_path)
        if self.is_spring_file_archived(spring_file_path):
            LOG.info(
                "Spring file already archived! Spring file will be removed from disk, continuing "
                "to next spring file "
            )
            self.remove_archived_spring_file(spring_file_path)
            return
        try:
            self.encryption_api.spring_symmetric_encryption(spring_file_path)
            self.encryption_api.key_asymmetric_encryption(spring_file_path)
            self.encryption_api.compare_spring_file_checksums(spring_file_path)
            self.pdc.archive_file_to_pdc(
                file_path=str(self.encryption_api.encrypted_spring_file_path(spring_file_path)),
                dry_run=self.dry_run,
            )
            self.pdc.archive_file_to_pdc(
                file_path=str(self.encryption_api.encrypted_key_path(spring_file_path)),
                dry_run=self.dry_run,
            )
            self.mark_file_as_archived(spring_file_path)
            self.encryption_api.cleanup(spring_file_path)
            self.remove_archived_spring_file(spring_file_path)
            LOG.debug(f"*** ARCHIVING PROCESS COMPLETED SUCCESSFULLY ***")
        except subprocess.CalledProcessError as error:
            LOG.error("Encryption failed: %s", error.stderr)
            LOG.debug(f"*** COMMAND PROCESS FAILED! ***")
            self.encryption_api.cleanup(spring_file_path)
        except ChecksumFailedError as error:
            LOG.error(error.message)
            self.encryption_api.cleanup(spring_file_path)
            LOG.debug(f"*** CHECKSUM PROCESS FAILED! ***")

    def retrieve_and_decrypt_spring_file(self, spring_file_path: Path) -> None:
        """Retrieves and decrypts a spring file and its decryption key"""
        LOG.info(f"*** START RETRIEVAL PROCESS OF SPRING FILE %s ***", spring_file_path)
        try:
            self.pdc.retrieve_file_from_pdc(
                file_path=str(self.encryption_api.encrypted_spring_file_path(spring_file_path)),
                dry_run=self.dry_run,
            )
            self.pdc.retrieve_file_from_pdc(
                file_path=str(self.encryption_api.encrypted_key_path(spring_file_path)),
                dry_run=self.dry_run,
            )
            self.encryption_api.key_asymmetric_decryption(spring_file_path)
            self.encryption_api.spring_symmetric_decryption(
                spring_file_path, output_file=spring_file_path
            )
        except subprocess.CalledProcessError as error:
            LOG.error("Decryption failed: %s", error.stderr)
            LOG.debug(f"*** RETRIEVAL PROCESS FAILED! ***")
        self.encryption_api.cleanup(spring_file_path)
        LOG.debug(f"*** RETRIEVAL PROCESS COMPLETED SUCCESSFULLY ***")

    def mark_file_as_archived(self, spring_file_path: Path) -> None:
        """Set the field 'to_archive' of the file in Housekeeper to mark that it has been
        archived to PDC"""
        if self.dry_run:
            LOG.info("Dry run, no changes made to %s", spring_file_path)
            return
        hk_spring_file: hk_models.File = self.hk_api.files(path=str(spring_file_path)).first()
        LOG.info("Setting %s to archived in Housekeeper", spring_file_path)
        self.hk_api.set_to_archive(file=hk_spring_file, value=True)

    def is_to_be_retrieved_and_decrypted(self, spring_file_path: Path) -> bool:
        """Determines if a spring file is archived on PDC and needs to be retrieved and decrypted"""
        return (
            self.hk_api.files(path=str(spring_file_path)).first().to_archive
            and not spring_file_path.exists()
        )

    def remove_archived_spring_file(self, spring_file_path: Path) -> None:
        """Removes all files related to spring PDC archiving"""
        if not self.dry_run:
            LOG.info("Removing spring file %s from disk", str(spring_file_path))
            spring_file_path.unlink()

    @property
    def dry_run(self) -> bool:
        """Dry run property"""
        LOG.debug("Backup dry run property: %s", self._dry_run)
        return self._dry_run

    @dry_run.setter
    def dry_run(self, value: bool) -> None:
        """Set the dry run property"""
        LOG.debug(
            "Set backup dry run property to %s, and set the encryption dry run property to the "
            "same value",
            value,
        )
        self._dry_run = value
        self.encryption_api.dry_run = value

    def is_spring_file_archived(self, spring_file_path: Path) -> bool:
        """Checks if a spring file is marked as archived in Housekeeper"""
        return self.hk_api.files(path=str(spring_file_path)).first().to_archive

    @staticmethod
    def is_compression_ongoing(spring_file_path: Path) -> bool:
        """Determines if (de)compression of the spring file ongoing"""
        return CompressionData(spring_file_path).pending_exists()



from pathlib import Path

import click
from cgmodels.cg.constants import StrEnum

SAMPLE_SHEET_HEADERS = {
    "bcl2fastq": [
        "FCID",
        "Lane",
        "SampleID",
        "SampleRef",
        "index",
        "index2",
        "SampleName",
        "Control",
        "Recipe",
        "Operator",
        "Project",
    ],
    "dragen": [
        "FCID",
        "Lane",
        "Sample_ID",
        "SampleRef",
        "index",
        "index2",
        "SampleName",
        "Control",
        "Recipe",
        "Operator",
        "Sample_Project",
    ],
}

SAMPLE_SHEET_DATA_HEADER = "[Data]"

DRAGEN_BARCODE_MISMATCH = ["BarcodeMismatchIndex1,#", "BarcodeMismatchIndex2,#"]

OPTION_BCL_CONVERTER = click.option(
    "-b",
    "--bcl-converter",
    type=click.Choice(["bcl2fastq", "dragen"]),
    default="bcl2fastq",
    help="Specify bcl conversion software. Choose between bcl2fastq and dragen. Default is "
    "bcl2fastq.",
)

FASTQ_FILE_SUFFIXES = [".fastq", ".gz"]

DEMUX_STATS_PATH = {
    "bcl2fastq": {
        "demultiplexing_stats": Path("Stats") / Path("DemultiplexingStats.xml"),
        "conversion_stats": Path("Stats") / Path("ConversionStats.xml"),
        "runinfo": None,
    },
    "dragen": {
        "demultiplexing_stats": Path("Reports") / Path("Demultiplex_Stats.csv"),
        "conversion_stats": Path("Reports") / Path("Demultiplex_Stats.csv"),
        "adapter_metrics_stats": Path("Reports") / Path("Adapter_Metrics.csv"),
        "runinfo": Path("Reports") / Path("RunInfo.xml"),
    },
}

DRAGEN_PASSED_FILTER_PCT = 100.00000


class DemultiplexingDirsAndFiles(StrEnum):
    """Demultiplexing related dirs and files"""

    SAMPLE_SHEET_FILE_NAME: str = "SampleSheet.csv"
    UNALIGNED_DIR_NAME: str = "Unaligned"



from pathlib import Path

from cg.models.demultiplex.demux_results import DemuxResults, LogfileParameters
from cg.models.demultiplex.flowcell import Flowcell


def test_demux_results_instance(demultiplexed_flowcell: Path, flowcell_object: Flowcell):
    # GIVEN the path to a demultiplexed flowcell and a flowcell object

    # WHEN instantiating a demux results object
    demux_results = DemuxResults(
        demux_dir=demultiplexed_flowcell, flowcell=flowcell_object, bcl_converter="bcl2fastq"
    )

    # THEN assert that the results dir is the unaligned dir
    assert demux_results.results_dir == demultiplexed_flowcell / "Unaligned"
    # THEN assert that the results dir exists
    assert demux_results.results_dir.exists()


def test_parse_log(bcl2fastq_demux_results: DemuxResults):
    # GIVEN som demux results

    # WHEN parsed log file parameters
    log_file_parameters: LogfileParameters = bcl2fastq_demux_results.get_logfile_parameters()

    # THEN assert that it was bcl2fastq that was run
    assert log_file_parameters.id_string == "bcl2fastq v2.20.0.422"



import pytest

from cg.meta.backup.backup import BackupApi
from cg.meta.backup.pdc import PdcAPI
from cg.models.cg_config import CGConfig


@pytest.fixture(name="backup_context")
def fixture_backup_context(cg_context: CGConfig) -> CGConfig:
    cg_context.meta_apis["backup_api"] = BackupApi(
        status=cg_context.status_db,
        pdc_api=PdcAPI(),
        root_dir=cg_context.backup.root.dict(),
    )
    return cg_context



import os.path
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Union, Optional

from pydantic import BaseModel, validator


class BalsamicConfigAnalysis(BaseModel):
    """Balsamic analysis model

    Attributes:
        case_id: case internal ID
        analysis_type: analysis type (single, paired or pon)
        sequencing_type: analysis sequencing type (wgs or targeted)
        BALSAMIC_version: BALSAMIC version used to produce the analysis
        config_creation_date: config creation timestamp
    """

    case_id: str
    analysis_type: str
    sequencing_type: str
    BALSAMIC_version: str
    config_creation_date: Union[datetime, str]


class BalsamicConfigSample(BaseModel):
    """Sample attributes used for BALSAMIC analysis

    Attributes:
        file_prefix: sample basename
        sample_name: sample internal ID
        type: sample type (tumor or normal)
    """

    file_prefix: str
    sample_name: str
    type: str


class BalsamicConfigReference(BaseModel):
    """Metadata of reference files

    Attributes:
        reference_genome: reference genome fasta file
        reference_genome_version: reference genome build version
        reference_access_date: date the reference was used for the analysis
    """

    reference_genome: Path
    reference_genome_version: Union[None, Path, str]
    reference_access_date: Union[datetime, Path]

    @validator("reference_genome_version", always=True)
    def validate_genome_version(cls, value: Path, values: dict) -> str:
        """
        Returns the genome version from the reference path:
        /home/proj/stage/cancer/balsamic_cache/X.X.X/hg19/genome/human_g1k_v37.fasta
        """

        return str(values["reference_genome"]).split("/")[-3]

    @validator("reference_access_date")
    def validate_reference_date(cls, value: Path) -> datetime:
        """Formats the reference date"""

        return datetime.strptime(os.path.basename(value), "%Y-%m-%d %H:%M:%S")


class BalsamicConfigPanel(BaseModel):
    """BALSAMIC attributes of a PANEL BED file if it's provided

    Attributes:
        capture_kit: string representation of a path to the PANEL BED file
        capture_kit_version: capture kit version
        chrom: list of chromosomes in PANEL BED
    """

    capture_kit: Path
    capture_kit_version: Optional[str]
    chrom: List[str]

    @validator("capture_kit_version", always=True)
    def validate_genome_version(cls, value: Path, values: dict) -> str:
        """
        Returns the panel bed version from the capture kit path:
        /home/proj/stage/cancer/reference/target_capture_bed/production/balsamic/gicfdna_3.1_hg19_design.bed
        """

        return os.path.basename(values["capture_kit"]).split("_")[-3]


class BalsamicConfigQC(BaseModel):
    """Config QC attributes

    Attributes:
        picard_rmdup: if the duplicates has been removed or not
        adapter: adapter sequence that has been trimmed
        quality_trim: whether quality trimming has been performed in the workflow
        adapter_trim: whether adapter trimming has been performed in the workflow
        umi_trim: whether UMI trimming has been performed in the workflow
        min_seq_length: minimum sequence length cutoff for reads
        umi_trim_length: UMI trimming length
    """

    picard_rmdup: bool
    adapter: Optional[str]
    quality_trim: bool
    adapter_trim: bool
    umi_trim: bool
    min_seq_length: Optional[str]
    umi_trim_length: Optional[str]


class BalsamicVarCaller(BaseModel):
    """Variant caller attributes model

    Attributes:
        mutation: if it addresses somatic or germline mutations
        type: called variant type (CNV, SV, SNV, etc.)
        analysis_type: analysis type supported by the variant caller (single, paired or pon)
        sequencing_type: sequencing type supported (wgs or targeted)
        workflow_solution: supported workflow (BALSAMIC, Sentieon, Sentieon_umi, etc.)
    """

    mutation: str
    type: str
    analysis_type: List[str]
    sequencing_type: List[str]
    workflow_solution: List[str]


class BalsamicConfigJSON(BaseModel):
    """Model for BALSAMIC analysis config validation

    Attributes:
        analysis: config analysis attributes
        samples: sample attributes associated to a specific case
        reference: BALSAMIC build reference
        panel: panel attributes (targeted analysis exclusively)
        umiworkflow: whether the UMI analysis was performed or not
    """

    analysis: BalsamicConfigAnalysis
    samples: Dict[str, BalsamicConfigSample]
    reference: BalsamicConfigReference
    panel: Optional[BalsamicConfigPanel]
    QC: BalsamicConfigQC
    vcf: Dict[str, BalsamicVarCaller]
    bioinfo_tools_version: Dict[str, List[str]]
    umiworkflow: bool



"""Tests the findbusinessdata part of the cg.store.api"""
from datetime import datetime

import pytest

from cg.store import Store, models
from cgmodels.cg.constants import Pipeline
from tests.store_helpers import StoreHelpers
from tests.store.api.conftest import fixture_rml_pool_store


def test_find_analysis_via_date(sample_store: Store, helpers: StoreHelpers):
    # GIVEN a case with an analysis with a startdate in the database
    analysis = helpers.add_analysis(store=sample_store, started_at=datetime.now())
    assert analysis.started_at

    # WHEN getting analysis via case_id and start date
    db_analysis = sample_store.analysis(analysis.family, analysis.started_at)

    # THEN the analysis should have been retrieved
    assert db_analysis == analysis


@pytest.mark.parametrize(
    "data_analysis",
    [Pipeline.BALSAMIC, Pipeline.BALSAMIC_UMI, Pipeline.MIP_DNA],
)
def test_families_by_subject_id(
    data_analysis: Pipeline,
    helpers: StoreHelpers,
    sample_store: Store,
):
    """Test that we get a case back for a subject id"""
    # GIVEN connected case exist
    subject_id = "a_subject_id"
    store: Store = sample_store

    rna_sample: models.Sample = helpers.add_sample(store=store, subject_id=subject_id)
    rna_case: models.Family = helpers.add_case(
        store=store, data_analysis=Pipeline.MIP_RNA, name="rna_case"
    )
    helpers.add_relationship(store=store, case=rna_case, sample=rna_sample)
    store.add_commit(rna_case)

    dna_sample: models.Sample = helpers.add_sample(store=store, subject_id=subject_id)
    dna_case: models.Family = helpers.add_case(
        store=store, data_analysis=data_analysis, name="dna_case"
    )
    helpers.add_relationship(store=store, case=dna_case, sample=dna_sample)
    store.add_commit(dna_case)

    customer: models.Customer = dna_case.customer

    # WHEN calling method families_by_subject_id
    all_cases: set[models.Family] = sample_store.families_by_subject_id(
        customer_id=customer.internal_id, subject_id=subject_id
    )

    dna_cases: set[models.Family] = sample_store.families_by_subject_id(
        customer_id=customer.internal_id, subject_id=subject_id, data_analyses=[data_analysis]
    )

    # THEN we got the case as result and another data_analysis not in result
    assert dna_case in all_cases
    assert rna_case in all_cases
    assert dna_case in dna_cases
    assert rna_case not in dna_cases


def test_families_by_subject_id_by_is_tumour(
    helpers: StoreHelpers,
    sample_store: Store,
):
    """Test that we get a case back for a subject id depending on tumour state"""
    # GIVEN we have two cases with same subject_id but different is_tumour
    subject_id = "a_subject_id"
    store: Store = sample_store

    tumour_sample: models.Sample = helpers.add_sample(
        store=store, subject_id=subject_id, is_tumour=True
    )
    tumour_case: models.Family = helpers.add_case(store=store, name="tumour_case")
    helpers.add_relationship(store=store, case=tumour_case, sample=tumour_sample)
    store.add_commit(tumour_case)

    non_tumour_sample: models.Sample = helpers.add_sample(
        store=store, subject_id=subject_id, is_tumour=False
    )
    non_tumour_case: models.Family = helpers.add_case(store=store, name="non_tumour_case")
    helpers.add_relationship(store=store, case=non_tumour_case, sample=non_tumour_sample)
    store.add_commit(non_tumour_case)

    customer: models.Customer = tumour_case.customer

    # WHEN calling method families_by_subject_id
    all_cases: set[models.Family] = sample_store.families_by_subject_id(
        customer_id=customer.internal_id, subject_id=subject_id
    )

    tumour_cases: set[models.Family] = sample_store.families_by_subject_id(
        customer_id=customer.internal_id, subject_id=subject_id, is_tumour=True
    )

    non_tumour_cases: set[models.Family] = sample_store.families_by_subject_id(
        customer_id=customer.internal_id, subject_id=subject_id, is_tumour=False
    )

    # THEN we have one group with both cases independent of is_tumour
    assert tumour_case in all_cases
    assert non_tumour_case in all_cases

    # THEN we have two groups depending on is_tumour
    assert tumour_case in tumour_cases
    assert tumour_case not in non_tumour_cases

    assert non_tumour_case in non_tumour_cases
    assert non_tumour_case not in tumour_cases


def test_get_latest_flow_cell_on_case(
    re_sequenced_sample_store: Store, case_id: str, flowcell_name: str
):
    """Test function to fetch the latest sequenced flowcell on a case"""

    # GIVEN a store with two flow cells in it, one being the latest sequenced of the two
    latest_flow_cell_obj: models.Flowcell = re_sequenced_sample_store.Flowcell.query.filter(
        models.Flowcell.name == flowcell_name
    ).first()

    # WHEN fetching the latest flow cell on a case with a sample that has been sequenced on both flow cells
    latest_flow_cell_on_case: models.Flowcell = (
        re_sequenced_sample_store.get_latest_flow_cell_on_case(family_id=case_id)
    )

    # THEN the fetched flow cell should have the same name as the other
    assert latest_flow_cell_obj.name == latest_flow_cell_on_case.name


def test_get_customer_id_from_ticket(analysis_store, customer_id, ticket_nr):
    """Tests if the function in fact returns the correct customer"""
    # Given a store with a ticket

    # Then the function should return the customer connected to the ticket
    assert analysis_store.get_customer_id_from_ticket(ticket_nr) == customer_id


def test_get_ticket_from_case(case_id: str, analysis_store_single_case, ticket_nr: int):
    """Tests if the correct ticket is returned for the given case"""
    # GIVEN a populated store with a case

    # WHEN the function is called
    ticket_from_case: int = analysis_store_single_case.get_ticket_from_case(case_id=case_id)

    # THEN the ticket should be correct
    assert ticket_nr == ticket_from_case


def test_get_case_pool(case_id: str, rml_pool_store: Store):
    """Tests if the correct pool is returned"""
    # GIVEN a case connected to a pool
    case_name = rml_pool_store.family(internal_id=case_id).name
    # WHEN the function is called
    pool = rml_pool_store.get_case_pool(case_id=case_id)
    # THEN the correct pool should be returned and the pool name should be the last part of the
    # case name
    assert pool.name == case_name.split("-", 1)[-1]



"""This script tests the cli methods to get analysis in status-db"""

from cg.cli.get import get
from cg.constants import RETURN_SUCCESS
from cg.models.cg_config import CGConfig
from cg.store import Store, models
from click.testing import CliRunner
from tests.store_helpers import StoreHelpers


def test_get_family_bad_case(cli_runner: CliRunner, base_context: CGConfig):
    """Test to get a analysis using a non-existing analysis-id"""
    # GIVEN an empty database

    # WHEN getting a analysis
    name = "dummy_name"
    result = cli_runner.invoke(get, ["family", name], obj=base_context)

    # THEN it should error about missing case instead of getting a case
    assert result.exit_code != RETURN_SUCCESS


def test_get_family_required(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test to get a case using only the required argument"""
    # GIVEN a database with a analysis
    family: models.Family = helpers.add_case(disk_store)
    internal_id = family.internal_id
    assert disk_store.Family.query.count() == 1

    # WHEN getting a analysis
    result = cli_runner.invoke(get, ["family", internal_id], obj=base_context)

    # THEN it should have been gotten
    assert result.exit_code == RETURN_SUCCESS
    assert internal_id in result.output



"""CLI function to compress FASTQ files into SPRING archives"""

import datetime as dt
import logging
from typing import Iterable, List, Optional

import click

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.cli.compress.helpers import correct_spring_paths, get_fastq_individuals, update_compress_api
from cg.constants.compression import CASES_TO_IGNORE
from cg.exc import CaseNotFoundError
from cg.meta.compress import CompressAPI
from cg.models.cg_config import CGConfig
from cg.store import Store, models

LOG = logging.getLogger(__name__)

# There is a list of problematic cases that we should skip


@click.command("fastq")
@click.option("-c", "--case-id", type=str)
@click.option("-n", "--number-of-conversions", default=5, type=int, show_default=True)
@click.option("-t", "--ntasks", default=12, show_default=True, help="Number of tasks for slurm job")
@click.option("-m", "--mem", default=50, show_default=True, help="Memory for slurm job")
@click.option(
    "-b",
    "--days-back",
    default=60,
    show_default=True,
    help="Threshold for how long ago family was created",
)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_obj
def fastq_cmd(
    context: CGConfig,
    case_id: Optional[str],
    number_of_conversions: int,
    ntasks: int,
    mem: int,
    days_back: int,
    dry_run: bool,
):
    """Find cases with FASTQ files and compress into SPRING"""
    LOG.info("Running compress FASTQ")
    compress_api: CompressAPI = context.meta_apis["compress_api"]
    store: Store = context.status_db
    update_compress_api(compress_api, dry_run=dry_run, ntasks=ntasks, mem=mem)

    cases: List[models.Family] = []
    if case_id:
        case_obj: models.Family = store.family(case_id)
        if not case_obj:
            LOG.warning("Could not find case %s", case_id)
            return
        cases.append(case_obj)
    else:
        date_threshold = dt.datetime.now() - dt.timedelta(days=days_back)
        cases = compress_api.get_cases_to_compress(store, date_threshold=date_threshold)

    case_conversion_count = 0
    ind_conversion_count = 0
    for case in cases:
        # Keeps track on if all samples in a case have been converted
        case_converted = True
        if case_conversion_count >= number_of_conversions:
            break
        internal_id: str = case.internal_id
        if internal_id in CASES_TO_IGNORE:
            LOG.info("Skipping case %s", internal_id)
            continue

        LOG.info("Searching for FASTQ files in case %s", internal_id)
        for link_obj in case.links:
            sample_id: str = link_obj.sample.internal_id
            case_converted: bool = compress_api.compress_fastq(sample_id)
            if not case_converted:
                LOG.info("skipping individual %s", sample_id)
                continue
            ind_conversion_count += 1
        if case_converted:
            case_conversion_count += 1

    LOG.info(
        "%s Individuals in %s (completed) cases where compressed",
        ind_conversion_count,
        case_conversion_count,
    )


@click.command("fastq")
@click.option("-c", "--case-id")
@click.option(
    "-b",
    "--days-back",
    default=60,
    show_default=True,
    help="Threshold for how long ago family was created",
)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_obj
def clean_fastq(context: CGConfig, case_id: Optional[str], days_back: int, dry_run: bool):
    """Remove compressed FASTQ files, and update links in housekeeper to SPRING files"""
    LOG.info("Running compress clean FASTQ")
    compress_api: CompressAPI = context.meta_apis["compress_api"]
    store: Store = context.status_db
    update_compress_api(compress_api, dry_run=dry_run)

    cases: List[models.Family] = []
    if case_id:
        case_obj: models.Family = store.family(case_id)
        if not case_obj:
            LOG.warning("Could not find case %s", case_id)
            return
        cases.append(case_obj)
    else:
        date_threshold = dt.datetime.now() - dt.timedelta(days=days_back)
        cases = compress_api.get_cases_to_compress(store, date_threshold=date_threshold)

    cleaned_inds = 0
    for case_obj in cases:
        if case_obj.internal_id in CASES_TO_IGNORE:
            continue
        samples: Iterable[str] = get_fastq_individuals(store=store, case_id=case_obj.internal_id)
        for sample_id in samples:
            res: bool = compress_api.clean_fastq(sample_id)
            if not res:
                LOG.info("Skipping individual %s", sample_id)
                continue
            cleaned_inds += 1

    LOG.info("Cleaned fastqs in %s individuals", cleaned_inds)


@click.command("fix-spring")
@click.option("-b", "--bundle-name")
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_obj
def fix_spring(context: CGConfig, bundle_name: Optional[str], dry_run: bool):
    """Check if bundle(s) have non existing SPRING files and correct these"""
    LOG.info("Running compress clean FASTQ")
    compress_api = context.meta_apis["compress_api"]
    update_compress_api(compress_api, dry_run=dry_run)
    hk_api: HousekeeperAPI = compress_api.hk_api
    correct_spring_paths(hk_api=hk_api, bundle_name=bundle_name, dry_run=dry_run)


@click.command("sample")
@click.argument("sample-id", type=str)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_obj
def decompress_sample(context: CGConfig, sample_id: str, dry_run: bool):
    """Decompress SPRING file, and include links to FASTQ files in housekeeper"""

    compress_api: CompressAPI = context.meta_apis["compress_api"]
    update_compress_api(compress_api, dry_run=dry_run)

    was_decompressed: bool = compress_api.decompress_spring(sample_id)
    if not was_decompressed:
        LOG.info(f"Skipping sample {sample_id}")
        return 0
    LOG.info(f"Decompressed sample {sample_id}")
    return 1


@click.command("case")
@click.argument("case-id", type=str)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def decompress_case(context: click.Context, case_id, dry_run):
    """Decompress SPRING file, and include links to FASTQ files in housekeeper"""

    store: Store = context.obj.status_db
    try:
        samples: Iterable[str] = get_fastq_individuals(store, case_id)
        decompressed_inds = 0
        for sample_id in samples:
            decompressed_count: int = context.invoke(
                decompress_sample, sample_id=sample_id, dry_run=dry_run
            )
            decompressed_inds += decompressed_count
    except CaseNotFoundError:
        return
    LOG.info(f"Decompressed spring archives in {decompressed_inds} samples")


@click.command("flowcell")
@click.argument("flowcell_id", type=str)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_obj
def decompress_flowcell(context: click.Context, flowcell_id: str, dry_run: bool):
    """Decompress SPRING file, and include links to FASTQ files in housekeeper"""

    store: Store = context.obj.status_db
    samples: Iterable[models.Sample] = store.get_samples_from_flowcell(flowcell_name=flowcell_id)
    decompressed_inds = 0
    for sample in samples:
        decompressed_count = context.invoke(
            decompress_sample, sample_id=sample.internal_id, dry_run=dry_run
        )
        decompressed_inds += decompressed_count
    LOG.info(f"Decompressed spring archives in {decompressed_inds} samples")


@click.command("ticket")
@click.argument("ticket_id", type=int)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def decompress_ticket(context: click.Context, ticket_id: int, dry_run: bool):
    """Decompress SPRING file, and include links to FASTQ files in housekeeper"""
    store: Store = context.obj.status_db
    samples: Iterable[models.Sample] = store.get_samples_from_ticket(ticket_id=ticket_id)
    decompressed_inds = 0
    for sample in samples:
        decompressed_count = context.invoke(
            decompress_sample, sample_id=sample.internal_id, dry_run=dry_run
        )
        decompressed_inds += decompressed_count
    LOG.info(f"Decompressed spring archives in {decompressed_inds} samples")



from typing import List, Optional
from pydantic import Field, validator

from cg.constants.orderforms import REV_SEX_MAP, SOURCE_TYPES
from cg.models.orders.sample_base import OrderSample


class ExcelSample(OrderSample):
    age_at_sampling: str = Field(None, alias="UDF/age_at_sampling")
    application: str = Field(..., alias="UDF/Sequencing Analysis")
    capture_kit: str = Field(None, alias="UDF/Capture Library version")
    collection_date: str = Field(None, alias="UDF/Collection Date")
    comment: str = Field(None, alias="UDF/Comment")
    concentration: str = Field(None, alias="UDF/Concentration (nM)")
    concentration_sample: str = Field(None, alias="UDF/Sample Conc.")
    container: str = Field(None, alias="Container/Type")
    container_name: str = Field(None, alias="Container/Name")
    control: str = Field(None, alias="UDF/Control")
    custom_index: str = Field(None, alias="UDF/Custom index")
    customer: str = Field(..., alias="UDF/customer")
    data_analysis: str = Field("MIP DNA", alias="UDF/Data Analysis")
    data_delivery: str = Field(None, alias="UDF/Data Delivery")
    elution_buffer: str = Field(None, alias="UDF/Sample Buffer")
    extraction_method: str = Field(None, alias="UDF/Extraction method")
    family_name: str = Field(None, alias="UDF/familyID")
    father: str = Field(None, alias="UDF/fatherID")
    formalin_fixation_time: str = Field(None, alias="UDF/Formalin Fixation Time")
    index: str = Field(None, alias="UDF/Index type")
    index_number: str = Field(None, alias="UDF/Index number")
    lab_code: str = Field(None, alias="UDF/Lab Code")
    mother: str = Field(None, alias="UDF/motherID")
    name: str = Field(..., alias="Sample/Name")
    organism: str = Field(None, alias="UDF/Strain")
    organism_other: str = Field(None, alias="UDF/Other species")
    original_lab: str = Field(None, alias="UDF/Original Lab")
    original_lab_address: str = Field(None, alias="UDF/Original Lab Address")
    panels: List[str] = Field(None, alias="UDF/Gene List")
    pool: str = Field(None, alias="UDF/pool name")
    post_formalin_fixation_time: str = Field(None, alias="UDF/Post Formalin Fixation Time")
    pre_processing_method: str = Field(None, alias="UDF/Pre Processing Method")
    priority: str = Field(None, alias="UDF/priority")
    quantity: str = Field(None, alias="UDF/Quantity")
    reagent_label: str = Field(None, alias="Sample/Reagent Label")
    reference_genome: str = Field(None, alias="UDF/Reference Genome Microbial")
    region: str = Field(None, alias="UDF/Region")
    region_code: str = Field(None, alias="UDF/Region Code")
    require_qcok: bool = Field(None, alias="UDF/Process only if QC OK")
    rml_plate_name: str = Field(None, alias="UDF/RML plate name")
    selection_criteria: str = Field(None, alias="UDF/Selection Criteria")
    sex: str = Field(None, alias="UDF/Gender")
    source: str = Field(None, alias="UDF/Source")
    status: str = Field(None, alias="UDF/Status")
    subject_id: str = Field(None, alias="UDF/subjectID")
    tissue_block_size: str = Field(None, alias="UDF/Tissue Block Size")
    tumour: bool = Field(None, alias="UDF/tumor")
    tumour_purity: str = Field(None, alias="UDF/tumour purity")
    volume: str = Field(None, alias="UDF/Volume (uL)")
    well_position: str = Field(None, alias="Sample/Well Location")
    well_position_rml: str = Field(None, alias="UDF/RML well position")

    @validator("data_analysis")
    def validate_data_analysis(cls, value):

        data_analysis_alternatives = [
            "Balsamic",  # OF 1508
            "Balsamic UMI",  # OF 1508
            "fastq",  # OF 1605
            "FLUFFY",  # OF 1604
            "MicroSALT",  # OF 1603 (implicit)
            "MIP DNA",  # OF 1508
            "MIP RNA",  # OF 1508
            "SARS-CoV-2",  # OF 2184
            "No analysis",  # OF 1508, 1604, 2184
        ]
        if value not in data_analysis_alternatives:
            raise AttributeError(f"'{value}' is not a valid data analysis")
        return value

    @validator("index_number", "volume", "quantity", "concentration", "concentration_sample")
    def numeric_value(cls, value: Optional[str]):
        if not value:
            return None
        str_value = value.rsplit(".0")[0]
        if str_value.replace(".", "").isnumeric():
            return str_value
        raise AttributeError(f"Order contains non-numeric value '{value}'")

    @validator("mother", "father")
    def validate_parent(cls, value: str):
        if value == "0.0":
            return None
        return value

    @validator("source")
    def validate_source(cls, value: Optional[str]):
        if value not in SOURCE_TYPES:
            raise ValueError(f"'{value}' is not a valid source")
        return value

    @validator("sex")
    def convert_sex(cls, value: Optional[str]):
        if not value:
            return None
        value = value.strip()
        return REV_SEX_MAP.get(value, "unknown")

    @validator("panels", pre=True)
    def parse_panels(cls, value):
        if not value:
            return None
        separator = ";"
        if ":" in value:
            separator = ":"
        return value.split(separator)

    @validator("data_delivery")
    def convert_data_delivery(cls, value: Optional[str]):
        return value.lower()

    @validator("status", "priority")
    def convert_to_lower(cls, value: Optional[str]):
        value = value.lower()
        return value

    @validator("priority")
    def convert_to_priority(cls, value: Optional[str]):
        if value.lower() == "förtur":
            return "priority"
        return value

    @validator("collection_date")
    def convert_to_date(cls, value: Optional[str]) -> Optional[str]:
        if not value:
            return None
        return value[0:10]



import json
from pathlib import Path
from typing import List

import pytest

from cg.apps.demultiplex.sample_sheet.index import Index
from cg.apps.demultiplex.sample_sheet.novaseq_sample_sheet import SampleSheetCreator
from cg.apps.lims.samplesheet import (
    LimsFlowcellSample,
    LimsFlowcellSampleBcl2Fastq,
    LimsFlowcellSampleDragen,
)
from cg.models.demultiplex.run_parameters import RunParameters


@pytest.fixture(name="demultiplexed_runs_dir")
def fixture_demultiplexed_runs_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to a dir with flowcells ready for demultiplexing"""
    return demultiplex_fixtures / "demultiplexed-runs"


@pytest.fixture(name="output_dirs_bcl2fastq")
def fixture_output_dirs_bcl2fastq(demultiplexed_runs_dir: Path) -> Path:
    """Return the output path a dir with flowcells that have finished demultiplexing using
    bcl2fastq"""
    return demultiplexed_runs_dir / "bcl2fastq"


@pytest.fixture(name="demux_run_dir")
def fixture_demux_run_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to a dir with flowcells ready for demultiplexing"""
    return demultiplex_fixtures / "flowcell-runs"


@pytest.fixture(name="demux_run_dir_bcl2fastq")
def fixture_demux_run_dir_bcl2fastq(demux_run_dir: Path) -> Path:
    """Return the path to a dir with flowcells ready for demultiplexing"""
    return demux_run_dir / "bcl2fastq"


@pytest.fixture(name="demux_run_dir_dragen")
def fixture_demux_run_dir_dragen(demux_run_dir: Path) -> Path:
    """Return the path to a dir with flowcells ready for demultiplexing"""
    return demux_run_dir / "dragen"


@pytest.fixture(name="index_obj")
def fixture_index_obj() -> Index:
    return Index(name="C07 - UDI0051", sequence="AACAGGTT-ATACCAAG")


@pytest.fixture(name="raw_samples_dir")
def fixture_raw_samples_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the raw samples fixtures"""
    return demultiplex_fixtures / "raw_lims_samples"


@pytest.fixture(name="novaseq_dir")
def fixture_novaseq_dir(demux_run_dir: Path, flowcell_full_name: str) -> Path:
    """Return the path to the novaseq demultiplex fixtures"""
    return demux_run_dir / flowcell_full_name


@pytest.fixture(name="flowcell_dir_bcl2fastq")
def fixture_novaseq_dir_bcl2fastq(demux_run_dir_bcl2fastq: Path, flowcell_full_name: str) -> Path:
    """Return the path to the novaseq demultiplex fixtures"""
    return demux_run_dir_bcl2fastq / flowcell_full_name


@pytest.fixture(name="flowcell_dir_dragen")
def fixture_novaseq_dir_dragen(demux_run_dir_dragen: Path, flowcell_full_name: str) -> Path:
    """Return the path to the novaseq demultiplex fixtures"""
    return demux_run_dir_dragen / flowcell_full_name


@pytest.fixture(name="hiseq_dir")
def fixture_hiseq_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the novaseq demultiplex fixtures"""
    return demultiplex_fixtures / "hiseq_run"


@pytest.fixture(name="unknown_run_parameters")
def fixture_unknown_run_parameters(demultiplex_fixtures: Path) -> Path:
    """Return the path to a file with hiseq run parameters"""
    return demultiplex_fixtures / "unknown_run_parameters.xml"


@pytest.fixture(name="run_parameters_missing_flowcell_type")
def fixture_run_parameters_missing_flowcell_type(demultiplex_fixtures: Path) -> Path:
    """Return the path to a file with hiseq run parameters"""
    return demultiplex_fixtures / "runParameters_missing_flowcell_run_field.xml"


@pytest.fixture(name="hiseq_run_parameters")
def fixture_hiseq_run_parameters(hiseq_dir: Path) -> Path:
    """Return the path to a file with hiseq run parameters"""
    return hiseq_dir / "runParameters.xml"


@pytest.fixture(name="novaseq_run_parameters")
def fixture_novaseq_run_parameters(novaseq_dir: Path) -> Path:
    """Return the path to a file with hiseq run parameters"""
    return novaseq_dir / "RunParameters.xml"


@pytest.fixture(name="raw_lims_sample")
def fixture_raw_lims_sample(flowcell_name: str) -> LimsFlowcellSample:
    """Return a raw lims sample"""
    sample = {
        "flowcell_id": flowcell_name,
        "lane": 1,
        "sample_id": "ACC7628A20",
        "sample_ref": "hg19",
        "index": "ACTGGTGTCG-ACAGGACTTG",
        "description": "",
        "sample_name": "814206",
        "control": "N",
        "recipe": "R1",
        "operator": "script",
        "project": "814206",
    }
    return LimsFlowcellSample(**sample)


@pytest.fixture(name="lims_novaseq_samples_file")
def fixture_lims_novaseq_samples_file(raw_samples_dir: Path) -> Path:
    """Return the path to a file with sample info in lims format"""
    return raw_samples_dir / "raw_samplesheet_novaseq.json"


@pytest.fixture(name="lims_novaseq_samples")
def fixture_lims_novaseq_samples(lims_novaseq_samples_file: Path) -> List[LimsFlowcellSample]:
    """Return a list of parsed flowcell samples"""
    with open(lims_novaseq_samples_file, "r") as in_file:
        raw_samples: List[dict] = json.load(in_file)
        return [LimsFlowcellSample(**sample) for sample in raw_samples]


@pytest.fixture(name="lims_novaseq_bcl2fastq_samples")
def fixture_lims_novaseq_bcl2fastq_samples(
    lims_novaseq_samples_file: Path,
) -> List[LimsFlowcellSampleBcl2Fastq]:
    """Return a list of parsed flowcell samples"""
    with open(lims_novaseq_samples_file, "r") as in_file:
        raw_samples: List[dict] = json.load(in_file)
        return [LimsFlowcellSampleBcl2Fastq(**sample) for sample in raw_samples]


@pytest.fixture(name="lims_novaseq_dragen_samples")
def fixture_lims_novaseq_dragen_samples(
    lims_novaseq_samples_file: Path,
) -> List[LimsFlowcellSampleDragen]:
    """Return a list of parsed flowcell samples"""
    with open(lims_novaseq_samples_file, "r") as in_file:
        raw_samples: List[dict] = json.load(in_file)
        return [LimsFlowcellSampleDragen(**sample) for sample in raw_samples]


@pytest.fixture(name="novaseq_run_parameters_object")
def fixture_novaseq_run_parameters_object(novaseq_run_parameters: Path) -> RunParameters:
    return RunParameters(novaseq_run_parameters)


@pytest.fixture(name="novaseq_bcl2fastq_sample_sheet_object")
def fixture_novaseq_bcl2fastq_sample_sheet_object(
    flowcell_name: str,
    lims_novaseq_bcl2fastq_samples: List[LimsFlowcellSampleBcl2Fastq],
    novaseq_run_parameters_object: RunParameters,
) -> SampleSheetCreator:
    return SampleSheetCreator(
        flowcell_id=flowcell_name,
        lims_samples=lims_novaseq_bcl2fastq_samples,
        run_parameters=novaseq_run_parameters_object,
        bcl_converter="bcl2fastq",
    )


@pytest.fixture(name="novaseq_dragen_sample_sheet_object")
def fixture_novaseq_dragen_sample_sheet_object(
    flowcell_name: str,
    lims_novaseq_dragen_samples: List[LimsFlowcellSampleDragen],
    novaseq_run_parameters_object: RunParameters,
) -> SampleSheetCreator:
    return SampleSheetCreator(
        flowcell_id=flowcell_name,
        lims_samples=lims_novaseq_dragen_samples,
        run_parameters=novaseq_run_parameters_object,
        bcl_converter="dragen",
    )
